@@ -1,6 +1,6 @@
 /* SPDX-License-Identifier: GPL-2.0 */
 /*
- * Copyright (C) 2015-2017 Jason A. Donenfeld <Jason@zx2c4.com>. All Rights Reserved.
+ * Copyright (C) 2015-2018 Jason A. Donenfeld <Jason@zx2c4.com>. All Rights Reserved.
  *
  * See doc/protocol.md for more info
  */
@@ -27,8 +27,8 @@ enum noise_lengths {
 #define noise_encrypted_len(plain_len) (plain_len + NOISE_AUTHTAG_LEN)
 
 enum cookie_values {
-	COOKIE_SECRET_MAX_AGE = 2 * 60 * HZ,
-	COOKIE_SECRET_LATENCY = 5 * HZ,
+	COOKIE_SECRET_MAX_AGE = 2 * 60,
+	COOKIE_SECRET_LATENCY = 5,
 	COOKIE_NONCE_LEN = XCHACHA20POLY1305_NONCELEN,
 	COOKIE_LEN = 16
 };
@@ -42,14 +42,14 @@ enum counter_values {
 enum limits {
 	REKEY_AFTER_MESSAGES = U64_MAX - 0xffff,
 	REJECT_AFTER_MESSAGES = U64_MAX - COUNTER_WINDOW_SIZE - 1,
-	REKEY_TIMEOUT = 5 * HZ,
-	REKEY_TIMEOUT_JITTER_MAX = HZ / 3,
-	REKEY_AFTER_TIME = 120 * HZ,
-	REJECT_AFTER_TIME = 180 * HZ,
-	INITIATIONS_PER_SECOND = HZ / 50,
+	REKEY_TIMEOUT = 5,
+	REKEY_TIMEOUT_JITTER_MAX_JIFFIES = HZ / 3,
+	REKEY_AFTER_TIME = 120,
+	REJECT_AFTER_TIME = 180,
+	INITIATIONS_PER_SECOND = 50,
 	MAX_PEERS_PER_DEVICE = 1U << 20,
-	KEEPALIVE_TIMEOUT = 10 * HZ,
-	MAX_TIMER_HANDSHAKES = (90 * HZ) / REKEY_TIMEOUT,
+	KEEPALIVE_TIMEOUT = 10,
+	MAX_TIMER_HANDSHAKES = 90 / REKEY_TIMEOUT,
 	MAX_QUEUED_INCOMING_HANDSHAKES = 4096, /* TODO: replace this with DQL */
 	MAX_STAGED_PACKETS = 128,
 	MAX_QUEUED_PACKETS = 1024 /* TODO: replace this with DQL */
@@ -60,8 +60,7 @@ enum message_type {
 	MESSAGE_HANDSHAKE_INITIATION = 1,
 	MESSAGE_HANDSHAKE_RESPONSE = 2,
 	MESSAGE_HANDSHAKE_COOKIE = 3,
-	MESSAGE_DATA = 4,
-	MESSAGE_TOTAL = 5
+	MESSAGE_DATA = 4
 };
 
 struct message_header {
