@@ -1,4 +1,7 @@
-/* Copyright (C) 2015-2017 Jason A. Donenfeld <Jason@zx2c4.com>. All Rights Reserved. */
+/* SPDX-License-Identifier: GPL-2.0
+ *
+ * Copyright (C) 2015-2018 Jason A. Donenfeld <Jason@zx2c4.com>. All Rights Reserved.
+ */
 
 #include "netlink.h"
 #include "device.h"
@@ -30,7 +33,7 @@ static const struct nla_policy peer_policy[WGPEER_A_MAX + 1] = {
 	[WGPEER_A_FLAGS]			= { .type = NLA_U32 },
 	[WGPEER_A_ENDPOINT]			= { .len = sizeof(struct sockaddr) },
 	[WGPEER_A_PERSISTENT_KEEPALIVE_INTERVAL]= { .type = NLA_U16 },
-	[WGPEER_A_LAST_HANDSHAKE_TIME]		= { .len = sizeof(struct timeval) },
+	[WGPEER_A_LAST_HANDSHAKE_TIME]		= { .len = sizeof(struct timespec) },
 	[WGPEER_A_RX_BYTES]			= { .type = NLA_U64 },
 	[WGPEER_A_TX_BYTES]			= { .type = NLA_U64 },
 	[WGPEER_A_ALLOWEDIPS]			= { .type = NLA_NESTED }
@@ -63,22 +66,20 @@ static struct wireguard_device *lookup_interface(struct nlattr **attrs, struct s
 
 struct allowedips_ctx {
 	struct sk_buff *skb;
-	unsigned int idx_cursor, idx;
+	unsigned int i;
 };
 
-static int get_allowedips(void *ctx, union nf_inet_addr ip, u8 cidr, int family)
+static int get_allowedips(void *ctx, const u8 *ip, u8 cidr, int family)
 {
 	struct nlattr *allowedip_nest;
 	struct allowedips_ctx *actx = ctx;
 
-	if (++actx->idx < actx->idx_cursor)
-		return 0;
-	allowedip_nest = nla_nest_start(actx->skb, actx->idx - 1);
+	allowedip_nest = nla_nest_start(actx->skb, actx->i++);
 	if (!allowedip_nest)
 		return -EMSGSIZE;
 
 	if (nla_put_u8(actx->skb, WGALLOWEDIP_A_CIDR_MASK, cidr) || nla_put_u16(actx->skb, WGALLOWEDIP_A_FAMILY, family) ||
-	    nla_put(actx->skb, WGALLOWEDIP_A_IPADDR, family == AF_INET6 ? sizeof(struct in6_addr) : sizeof(struct in_addr), &ip)) {
+	    nla_put(actx->skb, WGALLOWEDIP_A_IPADDR, family == AF_INET6 ? sizeof(struct in6_addr) : sizeof(struct in_addr), ip)) {
 		nla_nest_cancel(actx->skb, allowedip_nest);
 		return -EMSGSIZE;
 	}
@@ -87,9 +88,9 @@ static int get_allowedips(void *ctx, union nf_inet_addr ip, u8 cidr, int family)
 	return 0;
 }
 
-static int get_peer(struct wireguard_peer *peer, unsigned int index, unsigned int *allowedips_idx_cursor, struct sk_buff *skb)
+static int get_peer(struct wireguard_peer *peer, unsigned int index, struct allowedips_cursor *rt_cursor, struct sk_buff *skb)
 {
-	struct allowedips_ctx ctx = { .skb = skb, .idx_cursor = *allowedips_idx_cursor };
+	struct allowedips_ctx ctx = { .skb = skb };
 	struct nlattr *allowedips_nest, *peer_nest = nla_nest_start(skb, index);
 	bool fail;
 
@@ -102,14 +103,14 @@ static int get_peer(struct wireguard_peer *peer, unsigned int index, unsigned in
 	if (fail)
 		goto err;
 
-	if (!ctx.idx_cursor) {
+	if (!rt_cursor->seq) {
 		down_read(&peer->handshake.lock);
 		fail = nla_put(skb, WGPEER_A_PRESHARED_KEY, NOISE_SYMMETRIC_KEY_LEN, peer->handshake.preshared_key);
 		up_read(&peer->handshake.lock);
 		if (fail)
 			goto err;
 
-		if (nla_put(skb, WGPEER_A_LAST_HANDSHAKE_TIME, sizeof(struct timeval), &peer->walltime_last_handshake) || nla_put_u16(skb, WGPEER_A_PERSISTENT_KEEPALIVE_INTERVAL, peer->persistent_keepalive_interval / HZ) ||
+		if (nla_put(skb, WGPEER_A_LAST_HANDSHAKE_TIME, sizeof(struct timespec), &peer->walltime_last_handshake) || nla_put_u16(skb, WGPEER_A_PERSISTENT_KEEPALIVE_INTERVAL, peer->persistent_keepalive_interval) ||
 				nla_put_u64_64bit(skb, WGPEER_A_TX_BYTES, peer->tx_bytes, WGPEER_A_UNSPEC) || nla_put_u64_64bit(skb, WGPEER_A_RX_BYTES, peer->rx_bytes, WGPEER_A_UNSPEC))
 			goto err;
 
@@ -126,13 +127,12 @@ static int get_peer(struct wireguard_peer *peer, unsigned int index, unsigned in
 	allowedips_nest = nla_nest_start(skb, WGPEER_A_ALLOWEDIPS);
 	if (!allowedips_nest)
 		goto err;
-	if (routing_table_walk_ips_by_peer_sleepable(&peer->device->peer_routing_table, &ctx, peer, get_allowedips)) {
-		*allowedips_idx_cursor = ctx.idx;
+	if (allowedips_walk_by_peer(&peer->device->peer_allowedips, rt_cursor, peer, get_allowedips, &ctx, &peer->device->device_update_lock)) {
 		nla_nest_end(skb, allowedips_nest);
 		nla_nest_end(skb, peer_nest);
 		return -EMSGSIZE;
 	}
-	*allowedips_idx_cursor = 0;
+	memset(rt_cursor, 0, sizeof(*rt_cursor));
 	nla_nest_end(skb, allowedips_nest);
 	nla_nest_end(skb, peer_nest);
 	return 0;
@@ -149,9 +149,15 @@ static int get_device_start(struct netlink_callback *cb)
 
 	if (ret < 0)
 		return ret;
+	cb->args[2] = (long)kzalloc(sizeof(struct allowedips_cursor), GFP_KERNEL);
+	if (!cb->args[2])
+		return -ENOMEM;
 	wg = lookup_interface(attrs, cb->skb);
-	if (IS_ERR(wg))
+	if (IS_ERR(wg)) {
+		kfree((void *)cb->args[2]);
+		cb->args[2] = 0;
 		return PTR_ERR(wg);
+	}
 	cb->args[0] = (long)wg;
 	return 0;
 }
@@ -160,7 +166,8 @@ static int get_device_dump(struct sk_buff *skb, struct netlink_callback *cb)
 {
 	struct wireguard_device *wg = (struct wireguard_device *)cb->args[0];
 	struct wireguard_peer *peer, *next_peer_cursor = NULL, *last_peer_cursor = (struct wireguard_peer *)cb->args[1];
-	unsigned int peer_idx = 0, allowedips_idx_cursor = (unsigned int)cb->args[2];
+	struct allowedips_cursor *rt_cursor = (struct allowedips_cursor *)cb->args[2];
+	unsigned int peer_idx = 0;
 	struct nlattr *peers_nest;
 	bool done = true;
 	void *hdr;
@@ -173,7 +180,7 @@ static int get_device_dump(struct sk_buff *skb, struct netlink_callback *cb)
 	hdr = genlmsg_put(skb, NETLINK_CB(cb->skb).portid, cb->nlh->nlmsg_seq, &genl_family, NLM_F_MULTI, WG_CMD_GET_DEVICE);
 	if (!hdr)
 		goto out;
-	genl_dump_check_consistent(cb, hdr, &genl_family);
+	genl_dump_check_consistent(cb, hdr);
 
 	if (!last_peer_cursor) {
 		if (nla_put_u16(skb, WGDEVICE_A_LISTEN_PORT, wg->incoming_port) || nla_put_u32(skb, WGDEVICE_A_FWMARK, wg->fwmark) || nla_put_u32(skb, WGDEVICE_A_IFINDEX, wg->dev->ifindex) || nla_put_string(skb, WGDEVICE_A_IFNAME, wg->dev->name))
@@ -195,15 +202,16 @@ static int get_device_dump(struct sk_buff *skb, struct netlink_callback *cb)
 	ret = 0;
 	/* If the last cursor was removed via list_del_init in peer_remove, then we just treat
 	 * this the same as there being no more peers left. The reason is that seq_nr should
-	 * indicate to userspace that this isn't a coherent dump anyway, so they'll try again. */
+	 * indicate to userspace that this isn't a coherent dump anyway, so they'll try again.
+	 */
 	if (list_empty(&wg->peer_list) || (last_peer_cursor && list_empty(&last_peer_cursor->peer_list))) {
 		nla_nest_cancel(skb, peers_nest);
 		goto out;
 	}
 	lockdep_assert_held(&wg->device_update_lock);
 	peer = list_prepare_entry(last_peer_cursor, &wg->peer_list, peer_list);
-	list_for_each_entry_continue (peer, &wg->peer_list, peer_list) {
-		if (get_peer(peer, peer_idx++, &allowedips_idx_cursor, skb)) {
+	list_for_each_entry_continue(peer, &wg->peer_list, peer_list) {
+		if (get_peer(peer, peer_idx++, rt_cursor, skb)) {
 			done = false;
 			break;
 		}
@@ -224,42 +232,44 @@ out:
 	}
 	genlmsg_end(skb, hdr);
 	if (done) {
-		cb->args[1] = cb->args[2] = 0;
+		cb->args[1] = 0;
 		return 0;
 	}
 	cb->args[1] = (long)next_peer_cursor;
-	cb->args[2] = (long)allowedips_idx_cursor;
 	return skb->len;
 
 	/* At this point, we can't really deal ourselves with safely zeroing out
 	 * the private key material after usage. This will need an additional API
-	 * in the kernel for marking skbs as zero_on_free. */
+	 * in the kernel for marking skbs as zero_on_free.
+	 */
 }
 
 static int get_device_done(struct netlink_callback *cb)
 {
 	struct wireguard_device *wg = (struct wireguard_device *)cb->args[0];
 	struct wireguard_peer *peer = (struct wireguard_peer *)cb->args[1];
+	struct allowedips_cursor *rt_cursor = (struct allowedips_cursor *)cb->args[2];
 
 	if (wg)
 		dev_put(wg->dev);
+	kfree(rt_cursor);
 	peer_put(peer);
 	return 0;
 }
 
 static int set_port(struct wireguard_device *wg, u16 port)
 {
-	struct wireguard_peer *peer, *temp;
+	struct wireguard_peer *peer;
 
 	if (wg->incoming_port == port)
 		return 0;
-	socket_uninit(wg);
-	wg->incoming_port = port;
-	peer_for_each (wg, peer, temp, false)
+	list_for_each_entry(peer, &wg->peer_list, peer_list)
 		socket_clear_peer_endpoint_src(peer);
-	if (!netif_running(wg->dev))
+	if (!netif_running(wg->dev)) {
+		wg->incoming_port = port;
 		return 0;
-	return socket_init(wg);
+	}
+	return socket_init(wg, port);
 }
 
 static int set_allowedip(struct wireguard_peer *peer, struct nlattr **attrs)
@@ -274,9 +284,9 @@ static int set_allowedip(struct wireguard_peer *peer, struct nlattr **attrs)
 	cidr = nla_get_u8(attrs[WGALLOWEDIP_A_CIDR_MASK]);
 
 	if (family == AF_INET && cidr <= 32 && nla_len(attrs[WGALLOWEDIP_A_IPADDR]) == sizeof(struct in_addr))
-		ret = routing_table_insert_v4(&peer->device->peer_routing_table, nla_data(attrs[WGALLOWEDIP_A_IPADDR]), cidr, peer);
+		ret = allowedips_insert_v4(&peer->device->peer_allowedips, nla_data(attrs[WGALLOWEDIP_A_IPADDR]), cidr, peer, &peer->device->device_update_lock);
 	else if (family == AF_INET6 && cidr <= 128 && nla_len(attrs[WGALLOWEDIP_A_IPADDR]) == sizeof(struct in6_addr))
-		ret = routing_table_insert_v6(&peer->device->peer_routing_table, nla_data(attrs[WGALLOWEDIP_A_IPADDR]), cidr, peer);
+		ret = allowedips_insert_v6(&peer->device->peer_allowedips, nla_data(attrs[WGALLOWEDIP_A_IPADDR]), cidr, peer, &peer->device->device_update_lock);
 
 	return ret;
 }
@@ -307,7 +317,8 @@ static int set_peer(struct wireguard_device *wg, struct nlattr **attrs)
 		down_read(&wg->static_identity.lock);
 		if (wg->static_identity.has_identity && !memcmp(nla_data(attrs[WGPEER_A_PUBLIC_KEY]), wg->static_identity.static_public, NOISE_PUBLIC_KEY_LEN)) {
 			/* We silently ignore peers that have the same public key as the device. The reason we do it silently
-			 * is that we'd like for people to be able to reuse the same set of API calls across peers. */
+			 * is that we'd like for people to be able to reuse the same set of API calls across peers.
+			 */
 			up_read(&wg->static_identity.lock);
 			ret = 0;
 			goto out;
@@ -335,20 +346,23 @@ static int set_peer(struct wireguard_device *wg, struct nlattr **attrs)
 	if (attrs[WGPEER_A_ENDPOINT]) {
 		struct sockaddr *addr = nla_data(attrs[WGPEER_A_ENDPOINT]);
 		size_t len = nla_len(attrs[WGPEER_A_ENDPOINT]);
+
 		if ((len == sizeof(struct sockaddr_in) && addr->sa_family == AF_INET) || (len == sizeof(struct sockaddr_in6) && addr->sa_family == AF_INET6)) {
 			struct endpoint endpoint = { { { 0 } } };
+
 			memcpy(&endpoint.addr, addr, len);
 			socket_set_peer_endpoint(peer, &endpoint);
 		}
 	}
 
 	if (flags & WGPEER_F_REPLACE_ALLOWEDIPS)
-		routing_table_remove_by_peer(&wg->peer_routing_table, peer);
+		allowedips_remove_by_peer(&wg->peer_allowedips, peer, &wg->device_update_lock);
 
 	if (attrs[WGPEER_A_ALLOWEDIPS]) {
 		int rem;
 		struct nlattr *attr, *allowedip[WGALLOWEDIP_A_MAX + 1];
-		nla_for_each_nested (attr, attrs[WGPEER_A_ALLOWEDIPS], rem) {
+
+		nla_for_each_nested(attr, attrs[WGPEER_A_ALLOWEDIPS], rem) {
 			ret = nla_parse_nested(allowedip, WGALLOWEDIP_A_MAX, attr, allowedip_policy, NULL);
 			if (ret < 0)
 				goto out;
@@ -361,7 +375,8 @@ static int set_peer(struct wireguard_device *wg, struct nlattr **attrs)
 	if (attrs[WGPEER_A_PERSISTENT_KEEPALIVE_INTERVAL]) {
 		const u16 persistent_keepalive_interval = nla_get_u16(attrs[WGPEER_A_PERSISTENT_KEEPALIVE_INTERVAL]);
 		const bool send_keepalive = !peer->persistent_keepalive_interval && persistent_keepalive_interval && netif_running(wg->dev);
-		peer->persistent_keepalive_interval = (unsigned long)persistent_keepalive_interval * HZ;
+
+		peer->persistent_keepalive_interval = persistent_keepalive_interval;
 		if (send_keepalive)
 			packet_send_keepalive(peer);
 	}
@@ -391,9 +406,10 @@ static int set_device(struct sk_buff *skb, struct genl_info *info)
 	++wg->device_update_gen;
 
 	if (info->attrs[WGDEVICE_A_FWMARK]) {
-		struct wireguard_peer *peer, *temp;
+		struct wireguard_peer *peer;
+
 		wg->fwmark = nla_get_u32(info->attrs[WGDEVICE_A_FWMARK]);
-		peer_for_each (wg, peer, temp, false)
+		list_for_each_entry(peer, &wg->peer_list, peer_list)
 			socket_clear_peer_endpoint_src(peer);
 	}
 
@@ -408,26 +424,32 @@ static int set_device(struct sk_buff *skb, struct genl_info *info)
 
 	if (info->attrs[WGDEVICE_A_PRIVATE_KEY] && nla_len(info->attrs[WGDEVICE_A_PRIVATE_KEY]) == NOISE_PUBLIC_KEY_LEN) {
 		struct wireguard_peer *peer, *temp;
-		u8 public_key[NOISE_PUBLIC_KEY_LEN] = { 0 }, *private_key = nla_data(info->attrs[WGDEVICE_A_PRIVATE_KEY]);
+		u8 public_key[NOISE_PUBLIC_KEY_LEN], *private_key = nla_data(info->attrs[WGDEVICE_A_PRIVATE_KEY]);
+
 		/* We remove before setting, to prevent race, which means doing two 25519-genpub ops. */
-		bool unused __attribute((unused)) = curve25519_generate_public(public_key, private_key);
-		peer = pubkey_hashtable_lookup(&wg->peer_hashtable, public_key);
-		if (peer) {
-			peer_put(peer);
-			peer_remove(peer);
+		if (curve25519_generate_public(public_key, private_key)) {
+			peer = pubkey_hashtable_lookup(&wg->peer_hashtable, public_key);
+			if (peer) {
+				peer_put(peer);
+				peer_remove(peer);
+			}
 		}
+
+		down_write(&wg->static_identity.lock);
 		noise_set_static_identity_private_key(&wg->static_identity, private_key);
-		peer_for_each (wg, peer, temp, false) {
+		list_for_each_entry_safe(peer, temp, &wg->peer_list, peer_list) {
 			if (!noise_precompute_static_static(peer))
 				peer_remove(peer);
 		}
 		cookie_checker_precompute_device_keys(&wg->cookie_checker);
+		up_write(&wg->static_identity.lock);
 	}
 
 	if (info->attrs[WGDEVICE_A_PEERS]) {
 		int rem;
 		struct nlattr *attr, *peer[WGPEER_A_MAX + 1];
-		nla_for_each_nested (attr, info->attrs[WGDEVICE_A_PEERS], rem) {
+
+		nla_for_each_nested(attr, info->attrs[WGDEVICE_A_PEERS], rem) {
 			ret = nla_parse_nested(peer, WGPEER_A_MAX, attr, peer_policy, NULL);
 			if (ret < 0)
 				goto out;
@@ -448,10 +470,15 @@ out_nodev:
 	return ret;
 }
 
-static const struct genl_ops genl_ops[] = {
+#ifndef COMPAT_CANNOT_USE_CONST_GENL_OPS
+static const
+#else
+static
+#endif
+struct genl_ops genl_ops[] = {
 	{
 		.cmd = WG_CMD_GET_DEVICE,
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 14, 0)
+#ifndef COMPAT_CANNOT_USE_NETLINK_START
 		.start = get_device_start,
 #endif
 		.dumpit = get_device_dump,
@@ -466,12 +493,13 @@ static const struct genl_ops genl_ops[] = {
 	}
 };
 
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 10, 0)
-static struct genl_family genl_family __ro_after_init = {
+static struct genl_family genl_family
+#ifndef COMPAT_CANNOT_USE_GENL_NOPS
+__ro_after_init = {
 	.ops = genl_ops,
 	.n_ops = ARRAY_SIZE(genl_ops),
 #else
-static struct genl_family genl_family = {
+= {
 #endif
 	.name = WG_GENL_NAME,
 	.version = WG_GENL_VERSION,
@@ -480,12 +508,12 @@ static struct genl_family genl_family = {
 	.netnsok = true
 };
 
-int __init netlink_init(void)
+int __init genetlink_init(void)
 {
 	return genl_register_family(&genl_family);
 }
 
-void __exit netlink_uninit(void)
+void __exit genetlink_uninit(void)
 {
 	genl_unregister_family(&genl_family);
 }
